# Copyright (c) 2016, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

import logging

from django.middleware.csrf import CsrfViewMiddleware
from rest_framework.authentication import get_authorization_header

from .perms import find_rule
from .utils import get_current_app


LOGGER = logging.getLogger(__name__)


class RulesMiddleware(CsrfViewMiddleware):
    """
    Disables CSRF check if the HTTP request is forwarded.
    """

    def process_view(self, request, callback, callback_args, callback_kwargs):
        view_class = getattr(request.resolver_match.func, 'view_class', None)
        if hasattr(view_class, 'conditional_forward'):
            app = get_current_app(request)
            request.matched_rule, request.matched_params = find_rule(
                request, app)
            if (request.matched_rule and request.matched_rule.is_forward
                and request.method not in ('GET', 'HEAD', 'OPTIONS', 'TRACE')):
                # We are forwarding the request so the CSRF is delegated
                # to the application handling the forwarded request.
                #pylint:disable=protected-access
                request._dont_enforce_csrf_checks = True
                LOGGER.debug("dont enforce csrf checks on %s %s",
                    request.method, request.path)

        auth = get_authorization_header(request).split()
        if auth and auth[0].lower() in [b'basic', b'bearer']:
            # We need to support API calls from the command line.
            #pylint:disable=protected-access
            request._dont_enforce_csrf_checks = True
            LOGGER.debug("dont enforce csrf checks on %s %s because"\
                " we have an authorization header",
                request.method, request.path)

        return super(RulesMiddleware, self).process_view(
            request, callback, callback_args, callback_kwargs)



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Models for the rules application.
"""
from __future__ import unicode_literals

import datetime, json, logging
try:
    # Python 2
    from itertools import izip
except ImportError:
    # Python 3
    izip = zip # pylint:disable=invalid-name

from django.core.mail import get_connection as get_connection_base
from django.core.validators import RegexValidator
from django.db import models
from django.db.models import Q
from django.utils.encoding import python_2_unicode_compatible
from django.utils.timezone import utc
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from . import settings


LOGGER = logging.getLogger(__name__)


SUBDOMAIN_RE = r'^[-a-zA-Z0-9_]+\Z'
SUBDOMAIN_SLUG = RegexValidator(
    SUBDOMAIN_RE,
    _("Enter a valid subdomain consisting of letters, digits or hyphens."),
    'invalid'
)


@python_2_unicode_compatible
class Engagement(models.Model):
    """
    This model tracks engagement of a user with the website in order
    to answer questions such as:
        - Has a user either edited a page?
    """
    slug = models.SlugField()
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
        related_name='engagements')
    last_visited = models.DateTimeField(
        default=datetime.datetime(1971, 1, 1).replace(tzinfo=utc))
    # 1971 instead of 1970 to avoid Overflow exception in South.

    class Meta:
        unique_together = ('slug', 'user')

    def __str__(self):
        return "%s-%s" % (self.slug, self.user)


class AppManager(models.Manager):

    def get_or_create(self, defaults=None, **kwargs):
        app, created = super(AppManager, self).get_or_create(
            defaults=defaults, **kwargs)
        if created:
            for rank_min_one, (path, rule_op, is_forward) \
                in enumerate(settings.DEFAULT_RULES):
                #pylint:disable=no-member
                Rule.objects.db_manager(using=self._db).get_or_create(
                    app=app, rank=rank_min_one + 1,
                    defaults={'path':path, 'rule_op':rule_op,
                        'is_forward': is_forward})
        return app, created


@python_2_unicode_compatible
class BaseApp(models.Model): #pylint: disable=super-on-old-class
    """
    A ``App`` is used to select a database and a firewall profile.

    Implementation Note:
    Because we use separate databases, ``App`` and ``Organization``
    have no direct relationship (inheritance, foreign key, etc.).
    The same ``slug`` will though indistincly pickup a ``App`` and/or
    the matching ``Organization``.
    """
    NO_SESSION = 0
    COOKIE_SESSION_BACKEND = 1
    JWT_SESSION_BACKEND = 2

    SESSION_BACKEND_TYPE = (
        (NO_SESSION, "No session forwarded"),
        (COOKIE_SESSION_BACKEND, "Cookie based session backend"),
        (JWT_SESSION_BACKEND, "JWT based session backend"),
    )

    USER_REGISTRATION = 0
    PERSONAL_REGISTRATION = 1
    TOGETHER_REGISTRATION = 2
    IMPLICIT_REGISTRATION = 3

    REGISTRATION_TYPE = (
        (USER_REGISTRATION, "User registration"),
        (PERSONAL_REGISTRATION, "Personal registration"),
        (TOGETHER_REGISTRATION, "User and organization registration"),
        (IMPLICIT_REGISTRATION, "User registration wth implicit billing"),
    )

    objects = AppManager()

    # Since most DNS provider limit subdomain length to 25 characters,
    # we do here too.
    slug = models.SlugField(unique=True, max_length=25,
        validators=[SUBDOMAIN_SLUG], help_text=_(
            "unique identifier for the site (also serves as subdomain)"))

    account = models.ForeignKey(settings.ACCOUNT_MODEL,
        null=True, on_delete=models.CASCADE)

    # Fields for proxy features
    entry_point = models.URLField(max_length=100, null=True,
        help_text=_("Entry point to which requests will be redirected to"))
    enc_key = models.TextField(max_length=480, default="",
        verbose_name='Encryption Key',
        help_text=_("Encryption key used to sign proxyed requests"))
    session_backend = models.PositiveSmallIntegerField(
        choices=SESSION_BACKEND_TYPE, default=COOKIE_SESSION_BACKEND,
        help_text=_("Format to encode session in the forwarded HTTP request"))

    # XXX Fields used to custom signup form
    registration = models.PositiveSmallIntegerField(
        choices=REGISTRATION_TYPE, default=USER_REGISTRATION)

    class Meta:
        swappable = 'RULES_APP_MODEL'
        abstract = True

    def __str__(self): #pylint: disable=super-on-old-class
        return self.slug

    @property
    def printable_name(self): # XXX Organization full_name
        return str(self)

    def get_changes(self, update_fields):
        changes = {}
        for field_name in ('entry_point', 'enc_key',
                           'session_backend', 'registration'):
            pre_value = getattr(self, field_name, None)
            post_value = update_fields.get(field_name, None)
            if post_value is not None and pre_value != post_value:
                changes[field_name] = {'pre': pre_value, 'post': post_value}
        return changes

    def get_connection(self):
        #pylint:disable=no-self-use
        kwargs = {}
        return get_connection_base(**kwargs)

    def get_implicit_create_on_none(self):
        return self.registration in (
            self.PERSONAL_REGISTRATION, self.IMPLICIT_REGISTRATION)

    def get_from_email(self):
        #pylint:disable=no-self-use
        return settings.DEFAULT_FROM_EMAIL


@python_2_unicode_compatible
class App(BaseApp):

    def __str__(self): #pylint: disable=super-on-old-class
        return self.slug


class RuleManager(models.Manager):

    def get_rules(self, app, prefixes=None):
        """
        Get a list of access rules ordered by rank. When the optional
        *prefixes* parameter is specified, the list will be filtered
        such that any access rules returned start with one prefix present
        in the *prefixes* list.
        """
        args = []
        if prefixes is not None:
            for prefix in prefixes:
                if not args:
                    args = [Q(path__startswith=prefix)]
                else:
                    args[0] |= Q(path__startswith=prefix)
#            kwargs = {'path__startswith': prefixes}
        #pylint:disable=protected-access
        return self.db_manager(using=app._state.db).filter(
            *args, app=app).order_by('rank')


@python_2_unicode_compatible
class Rule(models.Model):
    """
    Rule to check in order to forward request, serve it locally
    and editing the content coming back through the proxy.
    """
    #pylint: disable=old-style-class,no-init

    HOME = 'index'
    ANY = 0

    objects = RuleManager()

    app = models.ForeignKey(settings.RULES_APP_MODEL,
        on_delete=models.CASCADE)
    # XXX At first I wanted to use a URLField for validation but this only
    #     works for URLs starting with a http/ftp protocol. What we really
    #     want here is a Path validator (including /).
    path = models.CharField(max_length=255,
        help_text=_("OpenAPI path against which requests are matched"))
    rule_op = models.PositiveSmallIntegerField(
        choices=settings.DB_RULE_OPERATORS, default=ANY,
        help_text=_("Method applied to grant or deny access"))
    kwargs = models.CharField(max_length=255, default="",
        help_text=_("Arguments to pass to the method to grant or deny access"))
    is_forward = models.BooleanField(default=True,
        help_text=_("When access is granted, should the request be forwarded"))
    engaged = models.CharField(max_length=50,
        help_text=_("Tags to check if it is the first time a user engages"))
    rank = models.IntegerField(
        help_text=_("Determines the order in which rules are considered"))
    moved = models.BooleanField(default=False)

    class Meta:
        unique_together = (('app', 'rank', 'moved'), ('app', 'path'))

    def __str__(self):
        return "%s/%s" % (self.app, self.path)

    def get_allow(self):
        if self.kwargs:
            return '%d/%s' % (self.rule_op, self.kwargs)
        return '%d' % self.rule_op

    def get_full_page_path(self):
        page_path = self.path
        path_prefix = ""
        if not page_path.startswith("/"):
            page_path = "/" + page_path
        if settings.PATH_PREFIX_CALLABLE:
            path_prefix = import_string(settings.PATH_PREFIX_CALLABLE)()
            if path_prefix and not path_prefix.startswith("/"):
                path_prefix = "/" + path_prefix
        return "%s%s" % (path_prefix, page_path)

    def match(self, request_path_parts):
        """
        Returns the dictionnary with paramaters in the request path
        or ``None`` if the Rule pattern does not match the request path.

        Example::

           Request path /profile/xia/ matched with /profile/:organization/
           will return {"organization": "xia"}.

           Request path /profile/xia/ matched with /billing/:organization/
           will return ``None``.
        """
        params = None
        page_path = self.get_full_page_path()
        # Normalize to avoid issues with path starting or ending with '/':
        pat_parts = [part for part in page_path.split('/') if part]
        if len(request_path_parts) >= len(pat_parts):
            # Only worth matching if the URL is longer than the pattern.
            try:
                params = json.loads(self.kwargs)
            except ValueError:
                params = {}
            for part, pat_part in izip(request_path_parts, pat_parts):
                if pat_part.startswith(':'):
                    slug = pat_part[1:]
                    if slug in params:
                        params.update({slug: part})
                elif part != pat_part:
                    return None
        return params



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json

from rest_framework.response import Response
from rest_framework.generics import RetrieveAPIView

from ..mixins import AppMixin, SessionDataMixin, UserMixin
from ..utils import JSONEncoder
from .serializers import SessionDataSerializer


class GetSessionDataRequest(object):

    def __init__(self, user):
        self.user = user


class GetSessionDataAPIView(SessionDataMixin, AppMixin, UserMixin,
                            RetrieveAPIView):
    """
    Returns a session data for a user as it will be passed to the backend
    service.

    **Examples

    .. code-block:: http

        GET /api/proxy/sessions/xia/ HTTP/1.1

    responds

    .. code-block:: json

        {
          "forward_session": "{username: xia}",
          "forward_session_header": "Authorization: XXX",
          "forward_url": "http://localhost:8001/"
        }
    """
    serializer_class = SessionDataSerializer

    def retrieve(self, request, *args, **kwargs):
        rule = None
        session = {}
        forward_session_header = self.get_forward_session_header(
            GetSessionDataRequest(self.user), self.app, rule, session)
        # Order of `forward_session_header =` and `forward_session =`
        # is important because of the call to `serialize_request`.
        forward_session = json.dumps(
            session, indent=2, cls=JSONEncoder)
        serializer = self.get_serializer({
            'forward_session': forward_session,
            'forward_session_header': forward_session_header,
            'forward_url': self.app.entry_point})
        return Response(serializer.data)



# Copyright (c) 2017, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime, json, logging

from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from django.utils.timezone import utc


LOGGER = logging.getLogger(__name__)


class JSONEncoder(json.JSONEncoder):

    def default(self, obj): #pylint: disable=method-hidden
        if hasattr(obj, 'isoformat'):
            return obj.isoformat()
        return super(JSONEncoder, self).default(obj)


def datetime_or_now(dtime_at=None):
    if not dtime_at:
        return datetime.datetime.utcnow().replace(tzinfo=utc)
    if dtime_at.tzinfo is None:
        dtime_at = dtime_at.replace(tzinfo=utc)
    return dtime_at


def get_app_model():
    """
    Returns the Site model that is active in this project.
    """
    from . import settings
    try:
        return django_apps.get_model(settings.RULES_APP_MODEL)
    except ValueError:
        raise ImproperlyConfigured(
            "RULES_APP_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured("RULES_APP_MODEL refers to model '%s'"\
" that has not been installed" % settings.RULES_APP_MODEL)


def get_current_app(request=None):
    """
    Returns the default app for a site.
    """
    from . import settings
    if settings.DEFAULT_APP_CALLABLE:
        app = import_string(settings.DEFAULT_APP_CALLABLE)(request=request)
        LOGGER.debug("rules.get_current_app: '%s'", app)
    else:
        app = get_app_model().objects.get(pk=settings.DEFAULT_APP_ID)
    return app



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.shortcuts import get_object_or_404
from django.utils import six

from .compat import reverse
from .utils import get_app_model, get_current_app


class AppMixinBase(object):
    """
    Returns a ``App`` from a URL.
    """
    app_url_kwarg = 'app'

    @property
    def app(self):
        if not hasattr(self, '_app'):
            if self.app_url_kwarg in self.kwargs:
                self._app = get_object_or_404(get_app_model(),
                    slug=self.kwargs.get(self.app_url_kwarg))
            else:
                self._app = get_current_app()
        return self._app

    def get_context_data(self, **kwargs):
        context = super(AppMixinBase, self).get_context_data(**kwargs)
        from . import settings
        account_url_kwarg = settings.ACCOUNT_URL_KWARG
        url_kwargs = {}
        if account_url_kwarg in self.kwargs:
            url_kwargs.update({
                account_url_kwarg: self.kwargs[account_url_kwarg]})
        urls = {
            'rules':{
               'api_rules': reverse('rules_api_rule_list', kwargs=url_kwargs),
               'api_detail': reverse('rules_api_app_detail', kwargs=url_kwargs),
               'api_generate_key': reverse(
                   'rules_api_generate_key', kwargs=url_kwargs),
               'api_session_data': reverse(
                   'rules_api_session_data_base', kwargs=url_kwargs),
               'app': reverse('rules_update', kwargs=url_kwargs)}}
        if 'urls' in context:
            for key, val in six.iteritems(urls):
                if key in context['urls']:
                    context['urls'][key].update(val)
                else:
                    context['urls'].update({key: val})
        else:
            context.update({'urls': urls})
        context.update({'app': self.app})
        return context



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

from deployutils.apps.django.backends.encrypted_cookies import (
    SessionStore as CookieSessionStore)
from deployutils.apps.django.backends.jwt_session_store import (
    SessionStore as JWTSessionStore)
from deployutils.apps.django.settings import SESSION_COOKIE_NAME
from django.contrib.auth import get_user_model
from django.utils.module_loading import import_string
from django.utils import six
from rest_framework.generics import get_object_or_404

from . import settings
from .compat import is_authenticated
from .models import Engagement
from .utils import datetime_or_now
from .extras import AppMixinBase


LOGGER = logging.getLogger(__name__)


class AppMixin(AppMixinBase, settings.EXTRA_MIXIN):
    pass


class EngagementMixin(object):
    """
    Records ``Engagement`` with a ``App``.
    """
    engagement_trigger = 'default'

    @property
    def last_visited(self):
        if not hasattr(self, '_last_visited'):
            self._last_visited = None
            if is_authenticated(self.request):
                engagement, created = Engagement.objects.get_or_create(
                    slug=self.engagement_trigger, user=self.request.user)
                if created:
                    # Avoid too many INSERT statements
                    engagement.last_visited = datetime_or_now()
                    engagement.save()
                    LOGGER.info("initial '%s' engagement with %s",
                        self.engagement_trigger, self.request.path)
                else:
                    self._last_visited = engagement.last_visited
        return self._last_visited

    def get_context_data(self, **kwargs):
        context = super(EngagementMixin, self).get_context_data(**kwargs)
        context.update({'last_visited': self.last_visited})
        return context


class SessionDataMixin(object):

    @staticmethod
    def serialize_request(request, app, rule):
        if is_authenticated(request):
            #pylint: disable=no-member
            serializer_class = import_string(settings.SESSION_SERIALIZER)
            serializer = serializer_class(request, context={
                'app': app, 'rule': rule})
            return serializer.data
        return {}

    @property
    def session_cookie_string(self):
        """
        Return the encrypted session information
        added to the HTTP Cookie Headers.
        """
        if not hasattr(self, '_session_cookie_string'):
            self._session_cookie_string = self.get_session_cookie_string(
                self.request, self.app, self.rule, self.session)
        return self._session_cookie_string

    def get_session_cookie_string(self, request, app, rule, session):
        # This is the latest time we can populate the session
        # since after that we need it to encrypt the cookie string.
        session.update(self.serialize_request(request, app, rule))
        session_store = CookieSessionStore(app.enc_key)
        session_token = session_store.prepare(session, app.enc_key)
        if not isinstance(session_token, six.string_types):
            # Because we don't want Python3 to prefix our strings with b'.
            session_token = session_token.decode('ascii')
        return session_token

    @property
    def session_jwt_string(self):
        """
        Return the encrypted session information
        encoded as a JWT token.
        """
        if not hasattr(self, '_session_jwt_string'):
            self._session_jwt_string = self.get_session_jwt_string(
                self.request, self.app, self.rule, self.session)
        return self._session_jwt_string

    def get_session_jwt_string(self, request, app, rule, session):
        # This is the latest time we can populate the session
        # since after that we need it to encrypt the cookie string.
        session.update(self.serialize_request(request, app, rule))
        session_store = JWTSessionStore(app.enc_key)
        session_token = session_store.prepare(session, app.enc_key)
        if not isinstance(session_token, six.string_types):
            # Because we don't want Python3 to prefix our strings with b'.
            #pylint:disable=redefined-variable-type
            session_token = session_token.decode('ascii')
        return session_token

    @property
    def forward_session_header(self):
        if not hasattr(self, '_forward_session_header'):
            if self.app.session_backend == self.app.JWT_SESSION_BACKEND:
                self._forward_session_header = "%s: %s" % (
                    SESSION_COOKIE_NAME, self.session_jwt_string)
            else:
                line = "%s: %s" % (SESSION_COOKIE_NAME,
                    self.session_cookie_string)
                self._forward_session_header = '\\\n'.join(
                    [line[i:i+48] for i in range(0, len(line), 48)])
        return self._forward_session_header

    def get_forward_session_header(self, request, app, rule, session):
        if app.session_backend == app.JWT_SESSION_BACKEND:
            forward_session_header = ("Authorization: Bearer %s" %
                self.get_session_jwt_string(request, app, rule, session))
        else:
            line = "Cookie: %s=%s" % (SESSION_COOKIE_NAME,
                self.get_session_cookie_string(request, app, rule, session))
            forward_session_header = '\\\n'.join(
                [line[i:i+48] for i in range(0, len(line), 48)])
        return forward_session_header


class UserMixin(object):

    user_field = 'username'
    user_url_kwarg = 'user'

    @property
    def user(self):
        if not hasattr(self, '_user'):
            slug = self.kwargs.get(self.user_url_kwarg)
            if getattr(self.request.user, self.user_field) == slug:
                # Not only do we avoid one database query, we also
                # make sure the user is the actual wrapper object.
                self._user = self.request.user
            else:
                kwargs = {self.user_field: slug}
                self._user = get_object_or_404(
                    get_user_model().objects.all(), **kwargs)
        return self._user



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
URLs for the resources API of rules.
"""

from django.conf.urls import url

from ... import settings
from ...api.keys import (AppUpdateAPIView, GenerateKeyAPIView)
from ...api.rules import RuleListAPIView, RuleDetailAPIView
from ...api.sessions import GetSessionDataAPIView

urlpatterns = [
    url(r'^proxy/sessions/(?P<user>%s)/?' % settings.USERNAME_PAT,
        GetSessionDataAPIView.as_view(), name='rules_api_session_data'),
    url(r'^proxy/sessions/?',
        GetSessionDataAPIView.as_view(), name='rules_api_session_data_base'),
    url(r'^proxy/key/',
        GenerateKeyAPIView.as_view(), name='rules_api_generate_key'),
    url(r'^proxy/rules(?P<rule>\S+)$',
        RuleDetailAPIView.as_view(), name='rules_api_rule_detail'),
    url(r'^proxy/rules',
        RuleListAPIView.as_view(), name='rules_api_rule_list'),
    url(r'^proxy/$',
        AppUpdateAPIView.as_view(), name='rules_api_app_detail'),
]



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import unicode_literals

import json

from django.utils import six
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from .. import settings
from ..models import Rule
from ..utils import get_app_model

#pylint: disable=no-init
#pylint: disable=old-style-class

class EnumField(serializers.Field):
    """
    Treat a ``PositiveSmallIntegerField`` as an enum.
    """
    choices = {}
    inverted_choices = {}

    def __init__(self, choices, *args, **kwargs):
        self.choices = dict(choices)
        self.inverted_choices = {
            val: key for key, val in six.iteritems(self.choices)}
        super(EnumField, self).__init__(*args, **kwargs)

    def to_representation(self, obj):
        if isinstance(obj, list):
            result = [self.choices.get(item, None) for item in obj]
        else:
            result = self.choices.get(obj, None)
        return result

    def to_internal_value(self, data):
        if isinstance(data, list):
            result = [self.inverted_choices.get(item, None) for item in data]
        else:
            result = self.inverted_choices.get(data, None)
        if result is None:
            if not data:
                raise serializers.ValidationError("This field cannot be blank.")
            raise serializers.ValidationError(
                "'%s' is not a valid choice. Expected one of %s." % (
                data, [choice for choice in six.itervalues(self.choices)]))
        return result


class NoModelSerializer(serializers.Serializer):

    def create(self, validated_data):
        raise RuntimeError('`create()` should not be called.')

    def update(self, instance, validated_data):
        raise RuntimeError('`update()` should not be called.')


class AppSerializer(serializers.ModelSerializer):

    class Meta:
        model = get_app_model()
        fields = ('slug', 'entry_point', 'session_backend')
        read_only_fields = ('slug',)

    @staticmethod
    def validate_entry_point(value):
        """
        Prevent unsafe URLs
        """
        parts = six.moves.urllib.parse.urlparse(value)
        if (not parts.netloc
            or parts.netloc.startswith('localhost')
            or parts.netloc.startswith('127.0.0.1')):
            raise serializers.ValidationError("Unsafe URLs are not allowed.")
        return value


class AppKeySerializer(serializers.ModelSerializer):
    """
    Used when a secret key is generated.
    """
    enc_key = serializers.CharField(read_only=True,
        help_text=_("Key used to decrypt the encoded session information."))

    class Meta:
        model = get_app_model()
        fields = ('enc_key',)


class RuleSerializer(serializers.ModelSerializer):

    allow = serializers.CharField(required=False, source='get_allow',
        help_text=_("Method applied to grant or deny access"))
    is_forward = serializers.BooleanField(required=False,
        help_text=_("When access is granted, should the request be forwarded"))
    engaged = serializers.CharField(required=False, allow_blank=True,
        help_text=_("Tags to check if it is the first time a user engages"))

    class Meta:
        model = Rule
        fields = ('rank', 'path', 'allow', 'is_forward', 'engaged')

    def validate(self, attrs):
        #pylint: disable=super-on-old-class,redefined-variable-type
        if 'get_allow' in attrs:
            parts = attrs.get('get_allow').split('/')
            try:
                rule_op = int(parts[0])
                if rule_op < 0 or rule_op >= len(settings.DB_RULE_OPERATORS):
                    raise ValueError("unknown rule %d" % rule_op)
                if len(parts) > 1:
                    params = json.loads(parts[1])
                    kwargs = {}
                    for key, dft in six.iteritems(
                            settings.RULE_OPERATORS[rule_op][2]):
                        if key in params:
                            kwargs[key] = params[key]
                        else:
                            kwargs[key] = dft
                    kwargs_encoded = json.dumps(kwargs)
                else:
                    kwargs_encoded = ""
                attrs['rule_op'] = rule_op
                attrs['kwargs'] = kwargs_encoded
            except ValueError as err:
                raise serializers.ValidationError(str(err))
        return super(RuleSerializer, self).validate(attrs)


class SessionDataSerializer(NoModelSerializer):

    forward_session = serializers.CharField(
        help_text=_("The session being forwarded"))
    forward_session_header = serializers.CharField(
        help_text=_("The HTTP header that encodes the session"))
    forward_url = serializers.CharField(
        help_text=_("The URL end point where the request is forwarded"))

    class Meta:
        fields = ('forward_session', 'forward_session_header', 'forward_url')
        read_only_fields = ('forward_session', 'forward_session_header',
            'forward_url')


class UsernameSerializer(NoModelSerializer):

    username = serializers.SerializerMethodField()

    class Meta:
        fields = ('username',)
        read_only_fields = ('username',)

    @staticmethod
    def get_username(request):
        return request.user.username



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint:disable=line-too-long
# There does not seem to be a way to make simple table with line breaks
# in cells.
"""
Convenience module for access of rules application settings,
which enforces default settings when the main settings module
does not contain the appropriate settings.

========================  ======================  =============
 Name                      Default                 Description
========================  ======================  =============
ACCOUNT_MODEL             AUTH_USER_MODEL         Model used in a multi-party implementation.
ACCOUNT_URL_KWARG         None                    Variable name used in url definition to select an account.
DEFAULT_APP_CALLABLE      None                    Function to get the default app.
DEFAULT_APP_ID            SITE_ID                 Primary key to get the default app.
DEFAULT_RULES             ('/', 0, False)         Rules used when creating a new account
EXTRA_MIXIN               object                  Mixin to derive from
PATH_PREFIX_CALLABLE      None                    Function to retrive the path prefix
RULE_OPERATORS            ('', 'login_required')  Rules that can be used to decorate a URL.
SESSION_SERIALIZER        ``UsernameSerializer``  Serializer used to represent sessions.
========================  ======================  =============

To override defaults, add a RULES configuration block to your project
settings.py

Example:

.. code-block:: python

    RULES = {
      'ACCOUNT_MODEL': 'saas.Organization'
    }

"""
from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils import inspect, six

from .compat import is_authenticated, reverse


_SETTINGS = {
    'ACCOUNT_MODEL': getattr(settings, 'AUTH_USER_MODEL', None),
    'ACCOUNT_URL_KWARG': None,
    'DEFAULT_APP_CALLABLE': None,
    'DEFAULT_APP_ID': getattr(settings, 'SITE_ID', 1),
    'DEFAULT_FROM_EMAIL': settings.DEFAULT_FROM_EMAIL,
    'DEFAULT_RULES': [('/', 0, False)],
    'EXTRA_MIXIN': object,
    'PATH_PREFIX_CALLABLE': None,
    'RULE_OPERATORS': (
        '',
        'rules.settings.fail_authenticated'),
    'SESSION_SERIALIZER': 'rules.api.serializers.UsernameSerializer'
}
_SETTINGS.update(getattr(settings, 'RULES', {}))

USERNAME_PAT = r'[\w.@+-]+'


def fail_authenticated(request):
    """
    Authenticated
    """
    if not is_authenticated(request):
        return reverse(settings.LOGIN_URL)
    return False


def _load_perms_func(path):
    """
    Load a function from its path as a string.
    """
    if not path:
        return "Any", None, {}
    if callable(path):
        func = path
    elif isinstance(path, six.string_types):
        dot_pos = path.rfind('.')
        module, attr = path[:dot_pos], path[dot_pos + 1:]
        try:
            mod = import_module(module)
        except ImportError as err:
            raise ImproperlyConfigured('Error importing %s: "%s"'
                % (path, err))
        except ValueError:
            raise ImproperlyConfigured('Error importing %s: ' % path)
        try:
            func = getattr(mod, attr)
        except AttributeError:
            raise ImproperlyConfigured('Module "%s" does not define a "%s"'
                % (module, attr))
    else:
        raise ImproperlyConfigured(
            '%s is neither callable or a module path' % path)

    short_name = ""
    for line in func.__doc__.splitlines():
        short_name = line.strip()
        if short_name:
            break
    if not short_name:
        raise ImproperlyConfigured(
            '%s should have at least a one-line documentation' % path)

    parms = {}
    for arg in inspect.get_func_full_args(func):
        if len(arg) > 1:
            parms.update({arg[0]: arg[1]})

    return short_name, func, parms


RULES_APP_MODEL = getattr(settings, 'RULES_APP_MODEL', 'rules.App')

AUTH_USER_MODEL = settings.AUTH_USER_MODEL
ACCOUNT_MODEL = _SETTINGS.get('ACCOUNT_MODEL')
ACCOUNT_URL_KWARG = _SETTINGS.get('ACCOUNT_URL_KWARG')
DEFAULT_APP_CALLABLE = _SETTINGS.get('DEFAULT_APP_CALLABLE')
DEFAULT_APP_ID = _SETTINGS.get('DEFAULT_APP_ID')
DEFAULT_FROM_EMAIL = _SETTINGS.get('DEFAULT_FROM_EMAIL')
DEFAULT_RULES = _SETTINGS.get('DEFAULT_RULES')
EXTRA_MIXIN = _SETTINGS.get('EXTRA_MIXIN')
PATH_PREFIX_CALLABLE = _SETTINGS.get('PATH_PREFIX_CALLABLE')
RULE_OPERATORS = tuple([_load_perms_func(item)
    for item in _SETTINGS.get('RULE_OPERATORS')])
SESSION_SERIALIZER = _SETTINGS.get('SESSION_SERIALIZER')

DB_RULE_OPERATORS = tuple([(idx, item[0])
    for idx, item in enumerate(RULE_OPERATORS)])

# We would use:
#     from django.core.validators import slug_re
#     SLUG_RE = slug_re.pattern
# but that includes ^ and $ which makes it unsuitable for use in URL patterns.
SLUG_RE = r'[a-zA-Z0-9_\-]+'



# Copyright (c) 2019, DjaoDjin inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json, logging, re

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.sites.requests import RequestSite
from django.core.exceptions import FieldError, SuspiciousOperation
from django.http import HttpResponse, SimpleCookie
from django.template.response import TemplateResponse
from django.utils import six
from django.views.generic import UpdateView, TemplateView
import requests
from requests.exceptions import RequestException
from deployutils.apps.django.settings import SESSION_COOKIE_NAME

from .. import settings
from ..compat import get_model
from ..mixins import AppMixin, SessionDataMixin
from ..perms import check_permissions as base_check_permissions
from ..utils import JSONEncoder, get_app_model

#pylint:disable=import-error
from django.utils.six.moves import http_cookies


LOGGER = logging.getLogger(__name__)


class DisallowedCookieName(SuspiciousOperation):
    """Cookie header contains invalid value"""
    pass


class SessionProxyMixin(SessionDataMixin):
    """
    Proxy to the application

    Check permissions associated to the request and forwards request
    when appropriate.
    """
    redirect_field_name = REDIRECT_FIELD_NAME
    login_url = None

    def check_permissions(self, request):
        redirect_url, self.rule, self.session = base_check_permissions(
            request, self.app,
            redirect_field_name=self.redirect_field_name,
            login_url=self.login_url)
        if self.session:
            # XXX Insert into self.request.session so we can use the same
            #     code on self-hosted templates.
            last_visited = self.session.get('last_visited', None)
            if last_visited:
                if not isinstance(last_visited, six.string_types):
                    last_visited = last_visited.isoformat()
                self.request.session.update({
                    'last_visited': last_visited})
        return (redirect_url, self.rule.is_forward if self.rule else False)

    def get_context_data(self, **kwargs):
        context = super(SessionProxyMixin, self).get_context_data(**kwargs)
        context.update({
            'forward_session': json.dumps(
                self.session, indent=2, cls=JSONEncoder),
            'forward_session_cookie': self.forward_session_header,
            'forward_url': '%s%s' % (self.app.entry_point, self.request.path),
        })
        return context

# Implementation Note:
# We cannot override ``dispatch()`` because djangorestframework happens
# to do user authentication there. Also doing so prevent injection
# of the edit tools.
# On the other hand, we cannnot blindly override `get`, `post`, etc. either
# because parents might not have implemented the method and it would result
# in 500 errors when incorrect HTTP requests are generated on the end points.
#   def dispatch(self, request, *args, **kwargs):
#       response = self.conditional_forward(request)
#       if response:
#           return response
#       # If we get here, the request was forwarded and we got an exception
#       # or the request must be handled locally.
#       return super(SessionProxyMixin, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        response = self.conditional_forward(request)
        if response:
            return response
        return self.forward_to_parent(request, *args, **kwargs)

    def forward_to_parent(self, request, *args, **kwargs):
        if request.method.lower() in self.http_method_names:
            handler = getattr(super(SessionProxyMixin, self),
                request.method.lower(), self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)

    def forward_error(self, err):
        context = self.get_context_data()
        context.update({'err': str(err)})
        return TemplateResponse(
            request=self.request,
            template='rules/forward_error.html',
            context=context,
            content_type='text/html',
            status=503)

    def conditional_forward(self, request):
        response, forward = self.check_permissions(request)
        if response:
            return response
        if forward:
            try:
                return self.fetch_remote_page()
            except RequestException as err:
                return self.forward_error(err)
        return None

    def fetch_remote_page(self):
        """
        Forward the request to the remote site after adjusting session
        information and request headers.

        Respond with the remote site response after adjusting session
        information and response headers.
        """
        remoteurl = '%s%s' % (self.app.entry_point, self.request.path)
        requests_args = self.translate_request_args(self.request)
        if LOGGER.getEffectiveLevel() == logging.DEBUG:
            LOGGER.debug("\"%s %s (Fwd to %s)\" with session %s,"\
                " updated headers: %s",
                self.request.method, self.request.path, self.app.entry_point,
                self.session, requests_args)
        else:
            LOGGER.info("\"%s %s (Fwd to %s)\"", self.request.method,
                self.request.path, self.app.entry_point, extra={
                    'event': 'http_forward', 'fwd_to': self.app.entry_point,
                    'request': self.request})
        response = requests.request(
            self.request.method, remoteurl, **requests_args)
        return self.translate_response(response)

    def translate_request_args(self, request):
        requests_args = {'allow_redirects': False, 'headers': {}}
        cookies = SimpleCookie()
        try:
            for key, value in six.iteritems(request.COOKIES):
                cookies[key] = value
            if (self.app.session_backend and
                self.app.session_backend != self.app.JWT_SESSION_BACKEND):
                cookies[SESSION_COOKIE_NAME] = self.session_cookie_string
        except http_cookies.CookieError as err:
            # Some is messing up with the 'Cookie' header. This sometimes
            # happen with bots trying to set 'Max-Age' or other reserved words.
            raise DisallowedCookieName(str(err))

        #pylint: disable=maybe-no-member
        # Something changed in `SimpleCookie.output` that creates an invalid
        # cookie string starting with a spacel or there are more strident
        # checks in the requests module (2.11.1) that prevents passing
        # a cookie string starting with a space.
        cookie_string = cookies.output(header='', sep=';').strip()

        # Retrieve the HTTP headers from a WSGI environment dictionary.  See
        # https://docs.djangoapp.com/en/dev/ref/request-response/\
        #    #django.http.HttpRequest.META
        headers = {}
        for key, value in six.iteritems(request.META):
            key_upper = key.upper()
            if key_upper.startswith('HTTP_'):
                key_upper = key_upper[5:].replace('_', '-')
                if key_upper == 'COOKIE':
                    headers[key_upper] = cookie_string
                else:
                    # Some servers don't like when you send the requesting host
                    # through but we do anyway. That's how the nginx reverse
                    # proxy is configured as well.
                    headers[key_upper] = value
            elif key_upper in ('CONTENT_TYPE', 'CONTENT_LENGTH'):
                headers[key.replace('_', '-')] = value

        site = RequestSite(request)
        if 'HOST' not in headers:
            headers.update({'HOST': site.domain})
        if 'X-FORWARDED-FOR' not in headers:
            headers.update({'X-FORWARDED-FOR': site.domain})
        if 'X-REAL-IP' not in headers:
            headers.update({'X-REAL-IP': request.META.get('REMOTE_ADDR', None)})
        if 'COOKIE' not in headers:
            headers.update({'COOKIE': cookie_string})

        if self.app.session_backend and \
            self.app.session_backend == self.app.JWT_SESSION_BACKEND:
            jwt_token = self.session_jwt_string
            headers.update({'AUTHORIZATION': 'Bearer %s' % jwt_token})

        if request.META.get(
                'CONTENT_TYPE', '').startswith('multipart/form-data'):
            if request.FILES:
                requests_args['files'] = request.FILES
            data = {}
            for key, val in six.iteritems(request.POST):
                data.update({key:val})
            requests_args['data'] = data
        else:
            requests_args['data'] = request.body

        params = request.GET.copy()

        # If there's a content-length header from Django, it's probably
        # in all-caps and requests might not notice it, so just remove it.
        for key in list(headers.keys()):
            if key.lower() == 'content-length':
                del headers[key]
            elif key.lower() == 'content-type' and request.META.get(
                'CONTENT_TYPE', '').startswith('multipart/form-data'):
                del headers[key]

        requests_args['headers'] = headers
        requests_args['params'] = params
        return requests_args

    @staticmethod
    def translate_response(response):
        proxy_response = HttpResponse(
            response.content, status=response.status_code)
        if 'set-cookie' in response.headers:
            # Here we have to decode the Set-Cookie ourselves because
            # requests will pack the set-cookie headers under the same key,
            # comma separated, which comma SimpleCookie.load() will append
            # to the path in the Morsel class (ie. Path=/,).
            # This of course results in the browser not sending the cookie
            # back to us later on.
            #pylint: disable=protected-access
            if six.PY2:
                set_cookie_lines \
                    = response.raw._original_response.msg.getallmatchingheaders(
                        'Set-Cookie')
            else:
                # We implement our own search here because
                # ``getallmatchingheaders`` is broken in Python3
                # (see https://bugs.python.org/issue5053)
                set_cookie_lines = []
                for line, data in six.iteritems(
                        response.raw._original_response.msg):
                    if line.lower() == 'set-cookie':
                        set_cookie_lines.append(line + ': ' + data)

            set_cookies_cont = ''
            set_cookies = []
            for line in set_cookie_lines:
                # The parsing is complicated by the fact that
                # ``getallmatchingheaders`` will return continuation lines
                # as an entry in the list.
                if line.startswith('Set-Cookie'):
                    if set_cookies_cont:
                        set_cookies += [set_cookies_cont]
                    set_cookies_cont = line[11:].strip()
                else:
                    set_cookies_cont += line.strip()
            if set_cookies_cont:
                set_cookies += [set_cookies_cont]

            intercepted_cookies = SimpleCookie()
            for cookie_text in set_cookies:
                intercepted_cookies.load(cookie_text)
            excluded_cookies = set([
                'sessionid', # XXX hardcoded
            ])
            for key in excluded_cookies:
                if key in intercepted_cookies:
                    del intercepted_cookies[key]
            #pylint: disable=maybe-no-member
            proxy_response.cookies.update(intercepted_cookies)

        excluded_headers = set([
            'set-cookie', # Previously parsed.
            # Hop-by-hop headers
            # ------------------
            # Certain response headers should NOT be just tunneled through.
            # For more info, see:
            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec13.html#sec13.5.1
            'connection', 'keep-alive', 'proxy-authenticate',
            'proxy-authorization', 'te', 'trailers', 'transfer-encoding',
            'upgrade',

            # Although content-encoding is not listed among the hop-by-hop
            # headers, it can cause trouble as well.  Just let the server set
            # the value as it should be.
            'content-encoding',

            # Since the remote server may or may not have sent the content
            # in the same encoding as Django will, let Django worry about
            # what the length should be.
            'content-length',
        ])
        for key, value in six.iteritems(response.headers):
            if key.lower() in excluded_headers:
                continue
            proxy_response[key] = value
        return proxy_response


class SessionProxyView(SessionProxyMixin, AppMixin, TemplateView):

    pass


class AppDashboardView(AppMixin, UpdateView):
    """
    Update a ``App``'s fields associated to the proxy dashboard
    (i.e. entry point and encryption key).
    """

    model = get_app_model()
    fields = ('entry_point',)
    template_name = 'rules/app_dashboard.html'

    def get_object(self, queryset=None):
        return self.app

    def get_context_data(self, **kwargs):
        #pylint:disable=too-many-locals
        context = super(AppDashboardView, self).get_context_data(**kwargs)
        rules = []
        for idx, rule in enumerate(settings.RULE_OPERATORS):
            short_name, _, defaults = rule
            parms = defaults.copy()
            look = re.search(r'%\((\S+)\)s', short_name)
            if look:
                cls_path = look.group(1)
                cls = get_model(cls_path)
                cls_name = cls_path.split('.')[-1].lower()
                try:
                    queryset = cls.objects.filter(
                        organization=self.app.account, is_active=True)
                except FieldError:
                    # The following code means we do not support optional
                    # ``organization`` foreign keys (i.e. ``null=True``).
                    try:
                        queryset = cls.objects.filter(
                            organization=self.app.account)
                    except FieldError:
                        queryset = cls.objects.all()
                for obj in queryset:
                    parms.update({cls_name: obj.slug})
                    identifier = '%d/%s' % (idx, json.dumps(parms))
                    parms.update()
                    descr = short_name % {cls_path: obj.title}
                    rules += [(identifier, descr)]
            else:
                rules += [(str(idx), short_name)]
        context.update({
            'site_available_at_url': self.request.build_absolute_uri('/'),
            'organization': self.app.account,
            'rules': rules
        })
        return context
